@@ -6,8 +6,8 @@ from .chem import *
 from .interp import *
 from .io import *
 from .land import *
+from .solvers import *
 from .spatial import *
 from .rad import *
 from .real import *
 from .thermo import *
-from .utils import *
\ No newline at end of file
