@@ -1,8 +1,9 @@
-from .emission_input import Emission_input
+from .emission_input import Emission_input, fit_gaussian_curve
 from .rfmip_background import get_rfmip_species
 from .tuv import calc_tuv_photolysis
 
 __all__ = [
     'Emission_input',
     'get_rfmip_species',
-    'calc_tuv_photolysis']
+    'calc_tuv_photolysis',
+    'fit_gaussian_curve']
